from abc import ABC, abstractmethod
from typing import Union, List, Tuple

import numpy as np

from .input import Keyboard, Mouse, Event
from .layer import LayerGroup, Layer


class GUIFacade(ABC):

    @property
    @abstractmethod
    def screenWidth(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def screenHeight(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def createWindow(self, title: str, width: int, height: int, fullScreen: bool = False):
        raise NotImplementedError()

    @abstractmethod
    def getTime(self) -> float:
        raise NotImplementedError()

    # Layer groups

    @abstractmethod
    def createLayerGroup(self, level: int = -1) -> LayerGroup:
        raise NotImplementedError()

    @abstractmethod
    def moveLayerGroup(self, layerGroup: Union[LayerGroup, int], level: int):
        raise NotImplementedError()

    @abstractmethod
    def removeLayerGroup(self, layerGroup: Union[LayerGroup, int]):
        raise NotImplementedError()

    @abstractmethod
    def layersChanged(self):
        raise NotImplementedError()

    # Main loop

    @property
    @abstractmethod
    def closingRequested(self) -> bool:
        raise NotImplementedError()

    @closingRequested.setter
    @abstractmethod
    def closingRequested(self, b: bool):
        raise NotImplementedError()

    @abstractmethod
    def init(self):
        raise NotImplementedError()

    @abstractmethod
    def updateInputs(self):
        raise NotImplementedError()

    @abstractmethod
    def render(self):
        raise NotImplementedError()

    @abstractmethod
    def quit(self):
        raise NotImplementedError()

    # Inputs

    @abstractmethod
    def getEvents(self) -> List[Event]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def keyboard(self) -> Keyboard:
        raise NotImplementedError()

    @property
    @abstractmethod
    def mouse(self) -> Mouse:
        raise NotImplementedError()

    # Capture

    @abstractmethod
    def captureScreenContent(self, x1: int, y1: int, x2: int, y2: int) -> np.ndarray:
        raise NotImplementedError()

    @abstractmethod
    def getPixelLayer(self, x: int, y: int) -> Tuple[Union[None, LayerGroup], int, Union[None, Layer], int]:
        raise NotImplementedError()


