from typing import List, Dict


class FaceAllocator:

    def __init__(self, step: int = 256):
        self.__step = step
        self.__face2id = [0] * step
        self.__ids = {
            0: [i for i in range(step)]
        }  # type: Dict[int, List[int]]

    @property
    def faceCount(self) -> int:
        return len(self.__face2id)

    @property
    def freeCount(self) -> int:
        return len(self.__ids[0])

    def allocateOne(self, faceId: int) -> int:
        if faceId <= 0:
            raise ValueError("Invalid id {}".format(faceId))
        freeIndices = self.__ids[0]

        # Increase capacity if the array is too small
        if len(freeIndices) == 0:
            oldLength = len(self.__face2id)
            newLength = len(self.__face2id) + self.__step
            freeIndices = [i for i in range(oldLength, newLength)]
            self.__face2id += [0] * self.__step
            self.__ids[0] = freeIndices

        # Remove one free index and add it to the faceId list
        freeIndex = freeIndices.pop(0)
        assert self.__face2id[freeIndex] == 0, "Internal error"
        self.__face2id[freeIndex] = faceId
        if faceId not in self.__ids:
            self.__ids[faceId] = [freeIndex]
        else:
            self.__ids[faceId].append(freeIndex)
        return freeIndex

    def allocate(self, faceId: int, count: int):
        if faceId <= 0:
            raise ValueError("Invalid id {}".format(faceId))
        freeIndices = self.__ids[0]

        # Increase capacity if the array is too small
        while len(freeIndices) < count:
            oldLength = len(self.__face2id)
            newLength = len(self.__face2id) + self.__step
            freeIndices += [i for i in range(oldLength, newLength)]
            self.__face2id += [0] * self.__step
            self.__ids[0] = freeIndices

        # Remove one free index and add it to the faceId list
        freeIndex = freeIndices[0:count]
        del freeIndices[0:count]
        for i in freeIndex:
            self.__face2id[i] = faceId
        if faceId not in self.__ids:
            self.__ids[faceId] = freeIndex
        else:
            self.__ids[faceId] += freeIndex

    def getIndices(self, faceId: int) -> List[int]:
        if faceId not in self.__ids:
            return []
        return self.__ids[faceId]

    def freeOne(self, index: int):
        if index < 0 or index >= len(self.__face2id):
            raise ValueError("Invalid index {}".format(index))
        faceId = self.__face2id[index]
        self.__face2id[index] = 0
        faceIndices = self.__ids[faceId]
        del faceIndices[faceIndices.index(index)]
        freeIndices = self.__ids[0]
        freeIndices.append(index)

    def free(self, faceId: int):
        if faceId <= 0:
            raise ValueError("Invalid id {}".format(faceId))
        if faceId not in self.__ids:
            return
        faceIndices = self.__ids[faceId]
        freeIndices = self.__ids[0]
        freeIndices += faceIndices
        for i in faceIndices:
            self.__face2id[i] = 0
        del self.__ids[faceId]

    def freeAll(self):
        self.__face2id = [0] * self.__step
        self.__ids = {
            0: [i for i in range(self.__step)]
        }

    def __str__(self):
        return "{} {}".format(self.__face2id, self.__ids)

    def check(self):
        ids = list(self.__ids.keys())
        for faceId in ids:
            checkIndices = []
            for i, id in enumerate(self.__face2id):
                if id == faceId:
                    checkIndices.append(i)
            indices = sorted(self.__ids[faceId])
            if indices != checkIndices:
                print(self)
                raise RuntimeError("Internal error {} {} != {}".format(faceId, indices, checkIndices))

    @staticmethod
    def test():
        faceAllocator = FaceAllocator(10)

        for repeat in range(2):
            for i in range(7):
                faceAllocator.allocateOne(1)
            assert faceAllocator.getIndices(1) == [0, 1, 2, 3, 4, 5, 6], faceAllocator
            faceAllocator.check()

            for i in range(7):
                faceAllocator.allocateOne(2)
            assert faceAllocator.getIndices(1) == [0, 1, 2, 3, 4, 5, 6], faceAllocator
            assert faceAllocator.getIndices(2) == [7, 8, 9, 10, 11, 12, 13], faceAllocator
            faceAllocator.check()

            faceAllocator.freeOne(2)
            assert faceAllocator.getIndices(1) == [0, 1, 3, 4, 5, 6], faceAllocator
            assert faceAllocator.getIndices(2) == [7, 8, 9, 10, 11, 12, 13], faceAllocator
            faceAllocator.check()

            for i in range(7):
                faceAllocator.allocateOne(1)
            assert faceAllocator.getIndices(1) == [0, 1, 3, 4, 5, 6, 14, 15, 16, 17, 18, 19, 2], faceAllocator
            assert faceAllocator.getIndices(2) == [7, 8, 9, 10, 11, 12, 13], faceAllocator
            faceAllocator.check()

            faceAllocator.free(2)
            assert faceAllocator.getIndices(1) == [0, 1, 3, 4, 5, 6, 14, 15, 16, 17, 18, 19, 2], faceAllocator
            assert faceAllocator.getIndices(2) == [], faceAllocator
            faceAllocator.check()

            faceAllocator.allocate(2, 9)
            assert faceAllocator.getIndices(1) == [0, 1, 3, 4, 5, 6, 14, 15, 16, 17, 18, 19, 2], faceAllocator
            assert faceAllocator.getIndices(2) == [7, 8, 9, 10, 11, 12, 13, 20, 21], faceAllocator
            faceAllocator.check()

            faceAllocator.freeAll()
            faceAllocator.check()




