import copy


class TextStyle:

    PLAIN = 0
    BOLD = 1
    ITALIC = 2
    UNDERLINE = 4
    SHADOW = 8
    OUTLINE = 16

    __slots__ = ['fontName', '__tileHeight', 'color', 'flags',
                 'underlineShift', 'underlineThickness',
                 'shadowShift', 'shadowColor',
                 'outlineThickness', 'outlineColor']

    def __init__(self, fontName: str, tileHeight: int, color: (int, int, int)):
        self.fontName = fontName
        self.__tileHeight = tileHeight
        self.color = color  # type: (int, int, int)
        self.flags = 0  # type: int
        self.shadowColor = (0, 0, 0)  # type: (int, int, int)
        self.outlineColor = (0, 0, 0)  # type: (int, int, int)
        self.underlineShift = 1  # type: int
        self.underlineThickness = 1  # type: int
        self.shadowShift = 1  # type: int
        self.outlineThickness = 1  # type: int
        self.updateShifts(tileHeight)

    @property
    def tileHeight(self) -> int:
        return self.__tileHeight

    @tileHeight.setter
    def tileHeight(self, tileHeight: int):
        self.__tileHeight = tileHeight

    def updateShifts(self, tileHeight: int):
        self.underlineShift = (tileHeight * 90) // 100
        self.underlineThickness = (tileHeight * 5) // 100
        if self.underlineThickness < 1:
            self.underlineThickness = 1

        self.shadowShift = tileHeight // 16
        if self.shadowShift < 1:
            self.shadowShift = 1

        self.outlineThickness = tileHeight // 32
        if self.outlineThickness < 1:
            self.outlineThickness = 1

    @property
    def italic(self) -> bool:
        return (self.flags & TextStyle.ITALIC) != 0

    @italic.setter
    def italic(self, enable: bool):
        if enable:
            self.flags |= TextStyle.ITALIC
        else:
            self.flags &= ~TextStyle.ITALIC

    @property
    def bold(self) -> bool:
        return (self.flags & TextStyle.BOLD) != 0

    @bold.setter
    def bold(self, enable: bool):
        if enable:
            self.flags |= TextStyle.BOLD
        else:
            self.flags &= ~TextStyle.BOLD

    @property
    def underline(self) -> bool:
        return (self.flags & TextStyle.UNDERLINE) != 0
    
    @underline.setter
    def underline(self, enable: bool):
        if enable:
            self.flags |= TextStyle.UNDERLINE
        else:
            self.flags &= ~TextStyle.UNDERLINE

    @property
    def shadow(self) -> bool:
        return (self.flags & TextStyle.SHADOW) != 0

    @shadow.setter
    def shadow(self, enable: bool):
        if enable:
            self.flags |= TextStyle.SHADOW
        else:
            self.flags &= ~TextStyle.SHADOW

    @property
    def outline(self) -> bool:
        return (self.flags & TextStyle.OUTLINE) != 0

    @outline.setter
    def outline(self, enable: bool):
        if enable:
            self.flags |= TextStyle.OUTLINE
        else:
            self.flags &= ~TextStyle.OUTLINE

    def __str__(self):
        description = "{} {} color:{}".format(self.fontName, self.tileHeight, self.color)
        if self.flags & TextStyle.ITALIC:
            description += " italic"
        if self.flags & TextStyle.BOLD:
            description += " bold"
        if self.flags & TextStyle.UNDERLINE:
            description += " underline:{} {}".format(self.underlineThickness, self.underlineShift)
        if self.flags & TextStyle.SHADOW:
            description += " shadow:{} {}".format(self.shadowShift, self.shadowColor)
        if self.flags & TextStyle.OUTLINE:
            description += " outline:{} {}".format(self.outlineThickness, self.outlineColor)
        return description

    def __eq__(self, other):
        return self.fontName == other.fontName \
            and self.tileHeight == other.tileHeight \
            and self.color == other.color \
            and self.flags == other.flags \
            and self.underlineShift == other.underlineShift \
            and self.underlineThickness == other.underlineThickness \
            and self.shadowShift == other.shadowShift \
            and self.shadowColor == other.shadowColor \
            and self.outlineThickness == other.outlineThickness \
            and self.outlineColor == other.outlineColor

    def __hash__(self):
        return hash(self.fontName) \
            + hash(self.tileHeight) \
            + hash(self.color) \
            + hash(self.flags) \
            + hash(self.underlineShift) \
            + hash(self.underlineThickness) \
            + hash(self.shadowShift) \
            + hash(self.shadowColor) \
            + hash(self.outlineThickness) \
            + hash(self.outlineColor)
