from abc import abstractmethod
from typing import Union

from gui import GUIFacade
from logic import Logic
from logic.command import MoveCharacterCommand
from render import RegionRenderer
from state import State, Region, StateListener
from .GameMode import GameMode


class DefaultGameMode(GameMode, StateListener):

    def __init__(self, gui: GUIFacade):
        self._gui = gui
        self._state = State()
        self._state.addListener(self)
        self._logic = Logic()

        self._viewX = 0  # type: int
        self._viewY = 0  # type: int
        self._currentRegionName = None  # type: Union[None, str]
        self._currentRegion = None  # type: Union[None, Region]

        self._tileSize = 32
        self._worldLayerGroup = gui.createLayerGroup()
        self._worldRenderer = RegionRenderer(gui, self._worldLayerGroup, self._tileSize)

    def updateState(self):
        if self._currentRegionName is not None:
            region = self._state.findRegion(self._currentRegionName)
            if region is not None:
                for characterIndex in range(region.getCharacterCount()):
                    self._logic.addCommand(MoveCharacterCommand(self._state, self._currentRegionName, characterIndex))
        self._logic.executeCommands()

    def render(self, time: float, epochFrame: float):
        self._worldRenderer.updateAnimations(time, epochFrame)
        self._gui.render()

    #############################################################
    # Events handling
    #############################################################

    def stateChanged(self, state: State):
        self._state = state
        self._state.addListener(self)
        self._currentRegionName = state.getRegionNames()[0]
        self.currentRegionChanged(self._state, self._currentRegionName)

    def currentRegionChanged(self, state: State, regionName: str):
        if regionName is None:
            self._currentRegionName = None
            self._currentRegion = None
            self._worldRenderer.updateLayer(None)
            self._gui.layersChanged()
            return

        self._currentRegionName = regionName
        region = state.getRegion(regionName)
        self._currentRegion = region
        self._worldRenderer.updateLayer(region)
        self._gui.layersChanged()
        self._worldRenderer.updateTiles(region)

    def regionCellChanged(self, state: State, regionName: str, x: int, y: int):
        if regionName != self._currentRegionName:
            return
        region = self._currentRegion
        for j in range(3):
            for i in range(3):
                xp = x + i - 1
                yp = y + j - 1
                if 0 <= xp < region.width and 0 <= yp < region.height:
                    self._worldRenderer.updateTile(region, xp, yp)

    def characterChanged(self, state: State, regionName: str, characterIndex: int):
        if regionName != self._currentRegionName:
            return
        region = self._currentRegion
        self._worldRenderer.updateCharacter(characterIndex, region.characters[characterIndex])
