import math
import random
from typing import Union

import typing

from gui import GUIFacade
from gui.layer import LayerGroup, GridLayer, CharactersLayer
from state import Region, State, Character
from state.Region import GROUND_ARRAY_INDEX, WATER_ARRAY_INDEX, TREE_ARRAY_INDEX
from .CharactersRenderer import CharactersRenderer
from .GroundRenderer import GroundRenderer
from .Renderer import Renderer
from .TreeRenderer import TreeRenderer
from .WaterRenderer import WaterRenderer


class RegionRenderer(Renderer):

    def __init__(self, gui: GUIFacade, layerGroup: LayerGroup, tileSize: int):
        super(RegionRenderer, self).__init__(gui)
        self.__layerGroup = layerGroup
        self.__tileSize = tileSize
        self.__regionWidth = 0
        self.__regionHeight = 0
        self.__groundRenderer = GroundRenderer(gui, layerGroup, 0, tileSize)
        self.__waterRenderer = WaterRenderer(gui, layerGroup, 1, tileSize)
        self.__treeRenderer = TreeRenderer(gui, layerGroup, 2, tileSize)
        self.__charactersRenderer = CharactersRenderer(gui, layerGroup, 3, 4, tileSize)
        self.__renderers = [
            self.__groundRenderer,
            self.__waterRenderer,
            self.__treeRenderer,
            self.__charactersRenderer
        ]

        # Earthquake effect
        self.__earthquakeType = ""
        self.__earthquakeFade = ""
        self.__earthquakeStart = 0.0
        self.__earthquakeEnd = 0.0

    def startEarthquake(self, typeName: str, fade: str, start: float, end: float):
        self.__earthquakeType = typeName
        self.__earthquakeFade = fade
        self.__earthquakeStart = start
        self.__earthquakeEnd = end

    def updateViewOnCharacter(self, characterIndex: int):
        # Compute character world location (pixel)
        worldX, worldY, worldZ = self.__charactersRenderer.getCharacterLocation(characterIndex)
        if worldX < 0 or worldY < 0:
            return
        worldX *= self.__tileSize
        worldY *= self.__tileSize

        # Update light location
        self.__layerGroup.setLightProperties(
            worldX + self.__tileSize / 2,
            worldY + self.__tileSize / 2,
            worldZ,
            2 * self.__tileSize)

        # Current view
        viewX, viewY = self.__layerGroup.getTranslation()
        screenX = worldX - viewX
        screenY = worldY - viewY

        # Move view if character is too close to the border of the screen
        margin = self._gui.screenWidth * 0.3
        if screenX < margin:
            viewX = worldX - margin
        elif screenX > self._gui.screenWidth - margin:
            viewX = worldX - self._gui.screenWidth + margin
        margin = self._gui.screenHeight * 0.3
        if screenY < margin:
            viewY = worldY - margin
        elif screenY > self._gui.screenHeight - margin:
            viewY = worldY - self._gui.screenHeight + margin

        # Only show world in the window
        if viewX < 0:
            viewX = 0
        if viewY < 0:
            viewY = 0
        maxX = self.__regionWidth * self.__tileSize - self._gui.screenWidth
        if viewX >= maxX:
            viewX = maxX
        maxY = self.__regionHeight * self.__tileSize - self._gui.screenHeight
        if viewY >= maxY:
            viewY = maxY

        self.__layerGroup.setTranslation(viewX, viewY)

    def updateLayer(self, region: Union[None, Region]):
        if region is not None:
            self.__regionWidth = region.width
            self.__regionHeight = region.height
        for renderer in self.__renderers:
            renderer.updateLayer(region)

    def updateTiles(self, region: Region):
        for renderer in self.__renderers:
            renderer.updateTiles(region)

    def updateTile(self, region: Region, x: int, y: int):
        for renderer in self.__renderers:
            renderer.updateTile(region, x, y)

    def updateCharacter(self, index: int, character: Character):
        for renderer in self.__renderers:
            renderer.updateCharacter(index, character)

    def updateAnimations(self, time: float, epochFrame: float):
        for renderer in self.__renderers:
            renderer.updateAnimations(time, epochFrame)

        # Night
        self.__layerGroup.setColorFactors(0.8, 0.6, 0.4)

        # Earthquake
        if self.__earthquakeType == "":
            self.__layerGroup.setShiftEffect(0, 0)
            return

        if time >= self.__earthquakeEnd:
            self.__earthquakeType = ""
            self.__layerGroup.setShiftEffect(0, 0)
            return

        f = 0
        if self.__earthquakeType == "random":
            f = random.uniform(-1, 1)
        elif self.__earthquakeType == "sinus":
            f = math.sin((time - self.__earthquakeStart) / 25)
        m = 1.0
        if self.__earthquakeFade == "linear":
            m = 1.0 - (time - self.__earthquakeStart) / (self.__earthquakeEnd - self.__earthquakeStart)
        x = f * m * 32
        self.__layerGroup.setShiftEffect(x, 0)


