from typing import List, Union

import numpy as np

from .Character import Character

REGION_MAX_WIDTH = 0x1000
REGION_MAX_HEIGHT = 0x1000

GROUND_ARRAY_INDEX = 0
GROUND_GRASS = 1
GROUND_SWAMP = 2
GROUND_DIRT = 3
GROUND_SAND = 4
GROUND_DEAD_LAND = 5
GROUND_MAX_VALUE = 5

WATER_ARRAY_INDEX = 1
WATER_NONE = 0
WATER_DARK = 1
WATER_RIVER = 2
WATER_SWAMP = 3
WATER_SEA = 4
WATER_CAVE = 5
WATER_MAX_VALUE = 5

TREE_ARRAY_INDEX = 2
TREE_NONE = 0
TREE_LIGHT_GREEN = 1
TREE_GREEN = 2
TREE_ORANGE = 3
TREE_DEAD = 4
TREE_STUMP = 5
TREE_STUMP_LARGE = 6
TREE_CAN_CROSS = 7
TREE_HERB_1 = 7
TREE_HERB_2 = 8
TREE_HERB_3 = 9
TREE_HERB_4 = 10
TREE_FLOWER_1 = 11
TREE_FLOWER_2 = 12
TREE_FLOWER_3 = 13
TREE_FLOWER_4 = 14
TREE_MUSHROOM_1 = 15
TREE_MUSHROOM_2 = 16
TREE_NENUPHAR_1 = 17
TREE_NENUPHAR_2 = 18
TREE_MAX_VALUE = 18


class Region:

    def __init__(self, width: int = 1, height: int = 1):
        self.__cells = np.zeros([width, height, 3], dtype=np.int8)
        self.__cells[:, :, GROUND_ARRAY_INDEX] = GROUND_GRASS
        self.__characters = []  # type: List[Union[None, Character]]

    @property
    def cells(self) -> np.ndarray:
        return self.__cells

    @cells.setter
    def cells(self, array: np.ndarray):
        assert array.ndim == 3
        assert array.shape[2] >= 3
        self.__cells = array

    @property
    def characters(self) -> List[Character]:
        return self.__characters

    @characters.setter
    def characters(self, array: List[Character]):
        self.__characters = array

    def getCharacterCount(self) -> int:
        return len(self.__characters)

    def findCharacter(self, index: int) -> Union[None, Character]:
        if 0 <= index < len(self.__characters):
            return self.__characters[index]
        return None

    def createCharacter(self, index: int, category: int):
        if self.findCharacter(index) is not None:
            raise ValueError("Character already exists")
        if index >= len(self.__characters):
            self.__characters.extend([None for l in range(len(self.__characters), index + 1)])
        character = Character(category)
        self.__characters[index] = character
        return character

    @property
    def width(self) -> int:
        return self.__cells.shape[0]

    @property
    def height(self) -> int:
        return self.__cells.shape[1]

    def getGroundValue(self, x: int, y: int) -> int:
        assert 0 <= x < self.width, "Invalid x = {}".format(x)
        assert 0 <= y < self.height, "Invalid y = {}".format(y)
        return self.__cells[x, y, GROUND_ARRAY_INDEX]

    def setGroundValue(self, x: int, y: int, value: int):
        assert 0 <= x < self.width
        assert 0 <= y < self.height
        assert 0 <= value <= GROUND_MAX_VALUE
        self.__cells[x, y, GROUND_ARRAY_INDEX] = value

    def getGroundNeightbors(self, x: int, y: int) -> np.ndarray:
        return self.getNeightbors(x, y, GROUND_ARRAY_INDEX)

    def getWaterValue(self, x: int, y: int) -> int:
        assert 0 <= x < self.width
        assert 0 <= y < self.height
        return self.__cells[x, y, WATER_ARRAY_INDEX]

    def setWaterValue(self, x: int, y: int, value: int):
        assert 0 <= x < self.width
        assert 0 <= y < self.height
        assert 0 <= value <= WATER_MAX_VALUE
        self.__cells[x, y, WATER_ARRAY_INDEX] = value

    def getWaterNeightbors(self, x: int, y: int) -> np.ndarray:
        return self.getNeightbors(x, y, WATER_ARRAY_INDEX)

    def getTreeValue(self, x: int, y: int) -> int:
        assert 0 <= x < self.width
        assert 0 <= y < self.height
        return self.__cells[x, y, TREE_ARRAY_INDEX]

    def setTreeValue(self, x: int, y: int, value: int):
        assert 0 <= x < self.width
        assert 0 <= y < self.height
        assert 0 <= value <= TREE_MAX_VALUE
        self.__cells[x, y, TREE_ARRAY_INDEX] = value

    def getTreeNeightbors(self, x: int, y: int) -> np.ndarray:
        return self.getNeightbors(x, y, TREE_ARRAY_INDEX)

    def getNeightbors(self, x: int, y: int, index: int) -> np.ndarray:
        assert 0 <= x < self.width
        assert 0 <= y < self.height
        width = self.width
        height = self.height
        if 0 < x < width - 1 and 0 < y < height - 1:
            # Common case, inside the map
            neightbors = self.__cells[x - 1:x + 2, y - 1:y + 2, index].copy()
        else:
            # Deal with corners
            neightbors = np.empty([3, 3], dtype=np.int8)
            neightbors.fill(self.__cells[x, y, index])
            if y > 0:
                if x > 0:
                    neightbors[0, 0] = self.__cells[x - 1, y - 1, index]
                neightbors[1, 0] = self.__cells[x, y - 1, index]
                if x < width - 1:
                    neightbors[2, 0] = self.__cells[x + 1, y - 1, index]
            if x > 0:
                neightbors[0, 1] = self.__cells[x - 1, y, index]
            if x < width - 1:
                neightbors[2, 1] = self.__cells[x + 1, y, index]
            if y < height - 1:
                if x > 0:
                    neightbors[0, 2] = self.__cells[x - 1, y + 1, index]
                neightbors[1, 2] = self.__cells[x, y + 1, index]
                if x < width - 1:
                    neightbors[2, 2] = self.__cells[x + 1, y + 1, index]

        return neightbors

    def canWalk(self, x: int, y: int) -> bool:
        if x < 0 or y < 0 or x >= self.width or y >= self.height:
            return False

        if self.getWaterValue(x, y) != WATER_NONE:
            return False

        return True
